export const CaseConvention = Object.freeze({
  Lowercase: 0,
  Uppercase: 1,
  PascalCase: 2,
  CamelCase: 3,
  SnakeCase: 4,
  ScreamingSnakeCase: 5,
  KebabCase: 6,
  ScreamingKebabCase: 7
} as const)

export type CaseConvention = typeof CaseConvention[keyof typeof CaseConvention]

const wordBoundaryRegex = /[A-Z]{2,}(?=[A-Z][a-z]+[0-9]*|\b)|[A-Z]?[a-z]+[0-9]*|[A-Z]|[0-9]+/g

function identifyWords(value: string) {
  return value && value.match(wordBoundaryRegex)
}

const lower = (ch: string) => ch.toLowerCase()
const upper = (ch: string) => ch.toUpperCase()

const first = <T>(xs: T[] | string) => xs && xs[0]
const tail = <T>(xs: T[] | string) => xs && xs.slice(1)

function upperFirst(xs: string) {
  return upper(first(xs)) + tail(xs)
}

function toPascalCase(words: string[]) {
  return words.map(lower).map(upperFirst).join('')
}

const joinMap = (fn: (x: string) => string, delim: string, xs: string[]) => {
  return xs.map(fn).join(delim)
}

export function convertCase(value: string, convention: CaseConvention) {
  const words = identifyWords(value)
  if (!words || words.length <= 0) {
    return ''
  }

  switch (convention) {
    case CaseConvention.Lowercase:
      return words.join('').toLowerCase()
    case CaseConvention.Uppercase:
      return words.join('').toUpperCase()
    case CaseConvention.PascalCase:
      return toPascalCase(words)
    case CaseConvention.CamelCase:
      const pascal = toPascalCase(words)
      return first<string>(pascal).toLowerCase() + tail(pascal)
    case CaseConvention.SnakeCase:
      return joinMap(lower, '_', words)
    case CaseConvention.ScreamingSnakeCase:
      return joinMap(upper, '_', words)
    case CaseConvention.KebabCase:
      return joinMap(lower, '-', words)
    case CaseConvention.ScreamingKebabCase:
      return joinMap(upper, '-', words)
  }
}
