import { Serde } from '../decorator'
import { SerdeOptions, Stage } from '../options'
import { ifNull, isFunction, isIterable, Nullable, orElse } from '../utils'
import { IterableSerializer, ObjectSerializer, Serializable, Serializer } from './interface'

const unhandledType = (serializer: any, value: any) => new TypeError(`'${serializer.constructor.name}' has no method for value type '${typeof value}'`)

function serializeEntries<T, K extends Serializable, V extends Serializable, E extends Iterable<[K, V]>>(serializer: ObjectSerializer<T>, value: E, options?: SerdeOptions) {
  let state

  for (const [key, val] of value) {
    const name = options?.getPropertyName(key as string, Stage.Serialize) ?? key
    state = serializer.serializeKey(name)
    state = serializer.serializeValue(val)
  }

  return serializer.end()
}

function serializeObject<T, K extends PropertyKey, V extends Serializable, R extends Record<K, V>>(serializer: ObjectSerializer<T>, value: R, options?: SerdeOptions) {
  return serializeEntries(serializer, Object.entries(value) as Iterable<[K, V]>, options)
}

function serializeIter<T, V extends Iterable<any>>(serializer: IterableSerializer<T>, value: V) {
  let state

  for (const val of value) {
    state = serializer.serializeElement(val)
  }

  return serializer.end()
}

function defaultOptions(value: any) {
  return value.constructor[Serde]
}

// dispatches in the order of serializeType -> serializeAny -> throw TypeError
export function serializeWith<T>(serializer: Serializer<T>, value: Serializable, optionsGetter: (value: any) => Nullable<SerdeOptions> = defaultOptions): Nullable<T> {
  // prepare fallback methods
  const serializeAny = orElse(
    serializer,
    serializer.serializeAny,
    (value: Serializable) => unhandledType(serializer, value)
  )

  const serialize = ifNull(serializer, serializeAny, value)

  switch (typeof value) {
    case 'string': return serialize(serializer.serializeString)
    case 'number': return serialize(serializer.serializeNumber)
    case 'bigint': return serialize(serializer.serializeBigInt)
    case 'boolean': return serialize(serializer.serializeBoolean)
    case 'symbol': return serialize(serializer.serializeSymbol)
    case 'undefined': return serialize(serializer.serializeNull)

    case 'object':
      if (isIterable(value) && isFunction(serializer.serializeIterable)) {
        return serializeIter(serializer.serializeIterable(), value)
      } else if (isFunction(serializer.serializeObject)) {
        return serializeObject(serializer.serializeObject!(), value as Record<PropertyKey, any>, optionsGetter(value))
      } // deliberate fallthrough when the above fail

    default: return serializeAny(value)
  }
}

